# -------------------------------------------------------------------------------
# Copyright (c) 2024-2025 Siemens
# All Rights Reserved.
# Authors: thomas.graf@siemens.com
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

from typing import Any, Dict, List, Optional

from .base import BaseMixin
from .sw360error import SW360Error


class PackagesMixin(BaseMixin):
    def get_package(self, package_id: str) -> Optional[Dict[str, Any]]:
        """Get information of about a package

        API endpoint: GET /package/{id}

        :param package_id: the id of the package to be requested
        :type package_id: string
        :return: a package
        :rtype: JSON package object
        :raises SW360Error: if there is a negative HTTP response
        """
        resp = self.api_get(self.url + "resource/api/packages/" + package_id)
        return resp

    def get_packages_by_name(self, name: str) -> List[Any]:
        """Gets a list of packages that match the given name.

        API endpoint: GET /packages?name=

        :param name: the name
        :type name: string
        :return: list of packages
        :rtype: list of JSON package objects
        :raises SW360Error: if there is a negative HTTP response
        """
        full_url = self.url + "resource/api/packages?name=" + name
        resp = self.api_get(full_url)
        if resp and ("_embedded" in resp) and ("sw360:packages" in resp["_embedded"]):
            return resp["_embedded"]["sw360:packages"]

        return []

    def get_all_packages(self, name: str = "", version: str = "", purl: str = "",
                         all_details: bool = False, page: int = -1,
                         page_size: int = -1, sort: str = "") -> Any:
        """Get information of about all packages

        API endpoint: GET /releases

        :param all_details: retrieve all package details (optional))
        :type all_details: bool
        :param page: page to retrieve
        :type page: int
        :param page_size: page size to use
        :type page_size: int
        :param sort: sort order for the packages ("name,desc"; "name,asc")
        :type sort: str
        :return: list of packages
        :rtype: list of JSON package objects
        :raises SW360Error: if there is a negative HTTP response
        """
        fullbase_url = self.url + "resource/api/packages"
        params = {}

        if all_details:
            params["allDetails"] = "true"

        if name:
            params["name"] = name

        if version:
            params["version"] = version

        if purl:
            params["purl"] = purl

        if page > -1:
            params["page"] = str(page)
            params["page_entries"] = str(page_size)

        if sort:
            params["sort"] = sort

        full_url = self._add_params(fullbase_url, params)
        resp = self.api_get(full_url)

        if page == -1 and resp and ("_embedded" in resp) and ("sw360:packages" in resp["_embedded"]):
            return resp["_embedded"]["sw360:packages"]

        return resp

    def get_packages_by_packagemanager(self, manager: str, page: int = -1,
                                       page_size: int = -1, sort: str = "") -> Any:
        """Get information of about all packages of a specific package manager

        API endpoint: GET /releases

        :param manager: name of the package manager
        :type manager: str
        :param page: page to retrieve
        :type page: int
        :param page_size: page size to use
        :type page_size: int
        :param sort: sort order for the packages ("name,desc"; "name,asc")
        :type sort: str
        :return: list of packages
        :rtype: list of JSON package objects
        :raises SW360Error: if there is a negative HTTP response
        """
        fullbase_url = self.url + "resource/api/packages"
        params = {"packageManager": manager}

        if page > -1:
            params["page"] = str(page)
            params["page_entries"] = str(page_size)

        if sort:
            params["sort"] = sort

        full_url = self._add_params(fullbase_url, params)
        resp = self.api_get(full_url)

        if page == -1 and resp and ("_embedded" in resp) and ("sw360:packages" in resp["_embedded"]):
            return resp["_embedded"]["sw360:packages"]

        return resp

    def create_new_package(self, name: str, version: str, purl: str,
                           package_type: str, package_details: Dict[str, Any] = {}) -> Optional[Dict[str, Any]]:
        """Create a new package

        API endpoint: POST /packages

        :param name: name of new package (usually set to component name)
        :param version: version string of new package (e.g. "1.0")
        :param purl: purl / package-url of the package
        :param package_type: CycloneDX package type of the package
        :param package_details: further package details as defined by SW360 REST API
        :type name: string
        :type version: string
        :type purl: string
        :type package_type: string
        :type package_details: dict
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        for param in "name", "version":
            package_details[param] = locals()[param]
        package_details["purl"] = purl
        package_details["packageType"] = package_type

        url = self.url + "resource/api/packages"
        response = self.api_post(
            url, json=package_details)
        if response is not None:
            if response.ok:
                return response.json()
        return None

    def update_package(self, package: Dict[str, Any], package_id: str) -> Optional[Dict[str, Any]]:
        """Update an existing package

        API endpoint: PATCH /packages

        :param release: the new package data
        :param release_id: the id of the package to be updated
        :type package: JSON
        :type package_id: string
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        if not package_id:
            raise SW360Error(message="No package id provided!")

        url = self.url + "resource/api/packages/" + package_id
        return self.api_patch(url, json=package)

    def delete_package(self, package_id: str) -> Optional[Dict[str, Any]]:
        """Delete an existing package

        API endpoint: DELETE /packages

        :param package_id: the id of the package to be deleted
        :type package_id: string
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        if not package_id:
            raise SW360Error(message="No package id provided!")

        url = self.url + "resource/api/packages/" + package_id
        response = self.api_delete(url)
        if response is not None:
            if response.ok:
                if response.text:
                    return response.json()
        return None



# -------------------------------------------------------------------------------
# Copyright (c) 2019-2025 Siemens
# Copyright (c) 2022 BMW CarIT GmbH
# All Rights Reserved.
# Authors: thomas.graf@siemens.com, gernot.hillier@siemens.com
# Authors: helio.chissini-de-castro@bmw.de
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

from typing import Any, Dict, List, Optional

from .base import BaseMixin
from .sw360error import SW360Error


class VendorMixin(BaseMixin):
    def get_all_vendors(self) -> List[Dict[str, Any]]:
        """Returns all vendors

        API endpoint: GET /vendors

        :return: a vendor
        :rtype: JSON vendor object
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(self.url + "resource/api/vendors")
        if resp and "_embedded" in resp and "sw360:vendors" in resp["_embedded"]:
            return resp["_embedded"]["sw360:vendors"]

        return []

    def get_vendor(self, vendor_id: str) -> Optional[Dict[str, Any]]:
        """Returns a vendor

        API endpoint: GET /vendors/{id}

        :param vendor_id: the id of the vendor to be requested
        :type vendor_id: string
        :return: list of vendors
        :rtype: list of JSON vendor objects
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(self.url + "resource/api/vendors/" + vendor_id)
        return resp

    def create_new_vendor(self, vendor: Dict[str, Any]) -> Dict[str, Any]:
        """Create a new vendor

        API endpoint: POST /vendors

        :param vendor: the new vendor data
        :type vendor: JSON vendor object
        :raises SW360Error: if there is a negative HTTP response
        """

        url = self.url + "resource/api/vendors"
        response = self.api_post(
            url, json=vendor)
        if response is not None:
            if response.ok:
                return response.json()
        raise SW360Error(response, url)

    def update_vendor(self, vendor: Dict[str, Any], vendor_id: str) -> Optional[Dict[str, Any]]:
        """Update an existing vendor

        API endpoint: PATCH /vendors

        :param vendor_id: the id of the vendor to be updated
        :type vendor_id: string
        :raises SW360Error: if there is a negative HTTP response
        """

        # 2019-04-03: error 405 - not allowed

        if not vendor_id:
            raise SW360Error(message="No vendor id provided!")

        url = self.url + "resource/api/vendors/" + vendor_id
        return self.api_patch(url, json=vendor)

    def delete_vendor(self, vendor_id: str) -> Optional[Dict[str, Any]]:
        """Delete an existing vendor

        API endpoint: DELETE /vendors

        :param vendor_id: the id of the vendor
        :type vendor_id: string
        :raises SW360Error: if there is a negative HTTP response
        """

        # 2019-04-03/2025-01-23: error 405 - not allowed

        if not vendor_id:
            raise SW360Error(message="No vendor id provided!")

        url = self.url + "resource/api/vendors/" + vendor_id

        response = self.api_delete(url)
        if response is not None:
            if response.ok:
                if response.text:
                    return response.json()

        return None



# -------------------------------------------------------------------------------
# Copyright (c) 2019-2025 Siemens
# Copyright (c) 2022 BMW CarIT GmbH
# All Rights Reserved.
# Authors: thomas.graf@siemens.com, gernot.hillier@siemens.com
# Authors: helio.chissini-de-castro@bmw.de
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

from typing import Any, Dict, List, Optional

from .base import BaseMixin
from .sw360error import SW360Error


class ComponentsMixin(BaseMixin):
    # return type List[Dict[str, Any]] | Optional[Dict[str, Any]] for Python 3.11 is good,
    # Union[List[Dict[str, Any]], Optional[Dict[str, Any]]] for lower Python versions is not good
    def get_all_components(self, fields: str = "", page: int = -1, page_size: int = -1,
                           all_details: bool = False,
                           sort: str = "") -> Any:
        """Get information of about all components

        API endpoint: GET /components

        :param page: page to retrieve
        :type page: int
        :param page_size: page size to use
        :type page_size: int
        :param all_details: retrieve all component details (optional))
        :type all_details: bool
        :param sort: sort order for the components ("name,desc"; "name,asc")
        :type sort: str
        :return: list of components
        :rtype: list of JSON component objects
        :raises SW360Error: if there is a negative HTTP response
        """

        fullbase_url = self.url + "resource/api/components"
        params = {}

        if all_details:
            params["allDetails"] = "true"

        if fields:
            params["fields"] = fields

        if page > -1:
            params["page"] = str(page)
            params["page_entries"] = str(page_size)

        if sort:
            params["sort"] = sort

        full_url = self._add_params(fullbase_url, params)
        resp = self.api_get(full_url)
        if not resp:
            return []

        if "_embedded" not in resp:
            return []

        if "sw360:components" not in resp["_embedded"]:
            return []

        if page == -1:
            return resp["_embedded"]["sw360:components"]

        return resp

    def get_components_by_type(
            self,
            component_type: str,
            page: int = -1,
            page_size: int = -1,
            sort: str = "") -> List[Dict[str, Any]]:
        """Get information of about all components for certain type

        API endpoint: GET /components

        :param component_type: the type of the component to be requested, one
         of INTERNAL, OSS, COTS, FREESOFTWARE, INNER_SOURCE, SERVICE
        :type component_type: string
        :param page: page to retrieve
        :type page: int
        :param page_size: page size to use
        :type page_size: int
        :param sort: sort order for the components ("name,desc"; "name,asc")
        :type sort: str
        :return: list of components
        :rtype: list of JSON component objects
        :raises SW360Error: if there is a negative HTTP response
        """

        fullbase_url = self.url + "resource/api/components"
        params = {"type": component_type}

        if page > -1:
            params["page"] = str(page)
            params["page_entries"] = str(page_size)

        if sort:
            params["sort"] = sort

        full_url = self._add_params(fullbase_url, params)
        resp = self.api_get(full_url)

        if resp and ("_embedded" in resp) and ("sw360:components" in resp["_embedded"]):
            return resp["_embedded"]["sw360:components"]

        return []

    def get_component(self, component_id: str) -> Optional[Dict[str, Any]]:
        """Get information of about a component

        API endpoint: GET /components/{id}

        :param component_id: the id of the component to be requested
        :type component_id: string
        :return: a component
        :rtype: JSON component object
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(self.url + "resource/api/components/" + component_id)
        return resp

    def get_component_by_url(self, component_url: str) -> Optional[Dict[str, Any]]:
        """Get information of about a component

        API endpoint: GET /components

        :param url: the full url of the component to be requested
        :type url: string
        :return: a component
        :rtype: JSON component object
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(component_url)
        return resp

    def get_component_by_name(
            self,
            component_name: str,
            page: int = -1,
            page_size: int = -1,
            sort: str = "") -> Optional[Dict[str, Any]]:
        """Get information of about a component

        API endpoint: GET /components

        :param component_name: the name of the component to look for
        :type component_name: string
        :param page: page to retrieve
        :type page: int
        :param page_size: page size to use
        :type page_size: int
        :param sort: sort order for the components ("name,desc"; "name,asc")
        :type sort: str
        :return: list of components
        :rtype: list of JSON component objects
        :raises SW360Error: if there is a negative HTTP response
        """

        fullbase_url = self.url + "resource/api/components"
        params = {"name": component_name}

        if page > -1:
            params["page"] = str(page)
            params["page_entries"] = str(page_size)

        if sort:
            params["sort"] = sort

        full_url = self._add_params(fullbase_url, params)
        resp = self.api_get(full_url)
        return resp

    def get_components_by_external_id(self, ext_id_name: str, ext_id_value: str = "") -> List[Dict[str, Any]]:
        """Get components by external id. `ext_id_value` can be left blank to
        search for all components with `ext_id_name`.

        API endpoint: GET /components

        :param ext_id_name: the name of the external id to look for
        :param ext_id_value: the value of the external id to look for
        :type ext_id_name: string
        :type ext_id_value: string
        :return: list of components
        :rtype: list of JSON component objects
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(
            self.url
            + "resource/api/components/searchByExternalIds?"
            + ext_id_name
            + "="
            + ext_id_value
        )
        if resp and ("_embedded" in resp) and ("sw360:components" in resp["_embedded"]):
            return resp["_embedded"]["sw360:components"]

        return []

    def create_new_component(self, name: str, description: str, component_type: str, homepage: str,
                             component_details: Dict[str, Any] = {}) -> Optional[Dict[str, Any]]:
        """Create a new component

        API endpoint: POST /components

        :param name: name of the new component
        :param description: description of the new component
        :param component_type: type of the new component, one of
         "INTERNAL", "OSS", "COTS", "FREESOFTWARE", "INNER_SOURCE", "SERVICE", "CODE_SNIPPET"
        :param homepage: homepage url of the new component
        :param component_details: further component details as defined by SW360 REST API
        :type name: string
        :type description: string
        :type component_type: string
        :type homepage: string
        :type component_details: dict
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        url = self.url + "resource/api/components"

        for param in "name", "description", "homepage":
            component_details[param] = locals()[param]
        component_details["componentType"] = component_type

        response = self.api_post(
            url, json=component_details)
        if response is not None:
            if response.ok:
                return response.json()
        return None

    def update_component(self, component: Dict[str, Any], component_id: str) -> Optional[Dict[str, Any]]:
        """Update an existing component

        API endpoint: PATCH /components

        :param component: the new component data
        :param component_id: the id of the component to be updated
        :type component: JSON component object
        :type component_id: string
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        if not component_id:
            raise SW360Error(message="No component id provided!")

        url = self.url + "resource/api/components/" + component_id
        return self.api_patch(url, json=component)

    def update_component_external_id(self, ext_id_name: str, ext_id_value: str,
                                     component_id: str, update_mode: str = "none") -> Optional[Dict[str, Any]]:
        """Set or update external id of a component. If the id is already set, it
        will only be changed if `update_mode=="overwrite"`. The id can be
        deleted using `update_mode=="delete"`.

        The method will return the old value of the external id or None if it
        was not set.

        API endpoint: PATCH /components

        :param ext_id_name: name of the external id
        :param ext_id_value: value of the external id
        :param component_id: the id of the component to be updated
        :param update_mode: can be "none" (default), "overwrite" or "delete"
        :type ext_id_name: string
        :type ext_id_value: string
        :type component_id: string
        :type update_mode: string
        :return: old value of external id
        :rtype: string
        :raises SW360Error: if there is a negative HTTP response
        """
        complete_data = self.get_component(component_id)
        if not complete_data:
            return None

        ret = self._update_external_ids(complete_data, ext_id_name,
                                        ext_id_value, update_mode)
        (old_value, data, update) = ret
        if update:
            self.update_component(data, component_id)
        return old_value

    def delete_component(self, component_id: str) -> Optional[Dict[str, Any]]:
        """Delete an existing component

        API endpoint: DELETE /components

        :param component_id: the id of the component to be deleted
        :type component_id: string
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        if not component_id:
            raise SW360Error(message="No component id provided!")

        url = self.url + "resource/api/components/" + component_id
        response = self.api_delete(url)
        if response is not None:
            if response.ok:
                if response.text:
                    return response.json()
        return None

    def get_users_of_component(self, component_id: str) -> Optional[Dict[str, Any]]:
        """Get information of about the users of a component

        API endpoint: GET /components/usedBy/{id}

        :param component_id: the id of the component to be requested
        :type component_id: string
        :return: all users of this component
        :rtype: JSON objects
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(self.url + "resource/api/components/usedBy/" + component_id)
        return resp

    def get_recent_components(self) -> Optional[List[Dict[str, Any]]]:
        """Get 5 of the service's most recently created components.

        API endpoint: GET /components/recentComponents

        :return: a list of components
        :rtype: JSON list of component objects
        :raises SW360Error: if there is a negative HTTP response
        """
        url = self.url + "resource/api/components/recentComponents"
        resp = self.api_get(url)
        if resp and ("_embedded" in resp) and ("sw360:components" in resp["_embedded"]):
            return resp["_embedded"]["sw360:components"]

        return []



# -------------------------------------------------------------------------------
# Copyright (c) 2023-2025 Siemens
# All Rights Reserved.
# Authors: thomas.graf@siemens.com
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlencode

import requests

from .sw360error import SW360Error


class BaseMixin():
    """Python interface to the Siemens SW360 platform

    Authentication against a running SW360 instance is performed using an API token.
    The token will be sent as HTTP header using the format
    `Authorization: <token_type> <token>`. Check your SW360 REST API
    documentation for details on needed type and how to get the token.
    token_type is "Bearer" for an OAuth workflow and "Token" for tokens
    generated via the SW360 UI.

    :param url: URL of the SW360 instance
    :param token: The SW360 REST API token (the cryptic string without
     "Authorization:" and `token_type`).
    :param oauth2: flag indicating whether this is an OAuth2 token
    :type url: string
    :type token: string
    :type oauth2: boolean
    """

    def __init__(self, url: str, token: str, oauth2: bool = False) -> None:
        """Constructor"""
        if url[-1] != "/":
            url += "/"
        self.url = url
        self.session: Optional[requests.Session] = None

        if oauth2:
            self.api_headers = {"Authorization": "Bearer " + token}
        else:
            self.api_headers = {"Authorization": "Token " + token}

        self.force_no_session = False

    def api_get(self, url: str = "") -> Optional[Dict[str, Any]]:
        """Request `url` from REST API and return json answer.

        :param url: the url to be requested
        :type url: string
        :return: JSON data
        :rtype: JSON
        :raises SW360Error: if there is a negative HTTP response
        """

        if (not self.force_no_session) and self.session is None:
            raise SW360Error(message="login_api needs to be called first")

        if self.force_no_session:
            response = requests.get(url, headers=self.api_headers)
        else:
            if self.session:
                response = self.session.get(url)

        if response.ok:
            if response.status_code == 204:  # 204 = no content
                return None
            return response.json()

        raise SW360Error(response, url)

    def api_post_multipart(self, url: str = "", files: Dict[str, Any] = {}) -> Optional[requests.Response]:
        """
        Send a multipart POST request to the specified URL with the provided file data.

        :param url: The URL to send the multipart POST request to.
        :type url: str
        :param files: The dictionary containing file data to be sent in the request.
        :type files: Dict[str, Any]
        :return: The JSON response received from the server, if any.
        :rtype: Optional[Dict[str, Any]]
        :raises SW360Error: If the HTTP response indicates an error.
        """

        if (not self.force_no_session) and self.session is None:
            raise SW360Error(message="login_api needs to be called first")

        if self.force_no_session:
            response = requests.post(url, headers=self.api_headers, files=files)
        else:
            if self.session:
                response = self.session.post(url, files=files)

        if response.ok:
            if response.status_code == 204:  # 204 = no content
                return None
            return response

        raise SW360Error(response, url)

    def api_post(
        self,
        url: str = "",
        json: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None
    ) -> Optional[requests.Response]:

        """
        Send a POST request to the specified URL with the provided json data.

        :param url: The URL to send the POST request to.
        :type url: str
        :param json: The dictionary containing json data to be sent in the request.
        :type json: Dict[str, Any]
        :return: The JSON response received from the server, if any.
        :rtype: Optional[Dict[str, Any]]
        :raises SW360Error: If the HTTP response indicates an error.
        """

        if (not self.force_no_session) and self.session is None:
            raise SW360Error(message="login_api needs to be called first")

        if self.force_no_session:
            response = requests.post(url, headers=self.api_headers, json=json)
        else:
            if self.session:
                response = self.session.post(url, json=json)

        if response.ok:
            if response.status_code == 204:  # 204 = no content
                return None
            return response

        raise SW360Error(response, url)

    def api_patch(self, url: str = "", json: Any = {}) -> Optional[Dict[str, Any]]:
        """
        Send a PATCH request to the specified URL with the provided json data.

        :param url: The URL to send the PATCH request to.
        :type url: str
        :param json: The dictionary containing json data to be sent in the request.
        :type json: Dict[str, Any]
        :return: The JSON response received from the server, if any.
        :rtype: Optional[Dict[str, Any]]
        :raises SW360Error: If the HTTP response indicates an error.
        """
        if (not self.force_no_session) and self.session is None:
            raise SW360Error(message="login_api needs to be called first")

        if self.force_no_session:
            response = requests.patch(url, headers=self.api_headers, json=json)
        else:
            if self.session:
                response = self.session.patch(url, json=json)

        if response.ok:
            if response.status_code == 204:  # 204 = no content
                return None
            if response.content:
                return response.json()
            else:
                return None

        raise SW360Error(response, url)

    def api_delete(self, url: str = "") -> Optional[requests.Response]:
        """Send a DELETE request to the specified `url` of the REST API and return JSON response.

        :param url: The URL to which the DELETE request will be sent.
        :type url: str
        :return: JSON data returned by the API, or None if the response is empty.
        :rtype: Optional[Dict[str, Any]]
        :raises SW360Error: If the API responds with a non-success HTTP status code.
        """
        if (not self.force_no_session) and self.session is None:
            raise SW360Error(message="login_api needs to be called first")

        if self.force_no_session:
            response = requests.delete(url, headers=self.api_headers)
        else:
            if self.session:
                response = self.session.delete(url)

        if response.ok:
            if response.status_code == 204:  # 204 = no content
                return None
            return response

        raise SW360Error(response, url)

    def _update_external_ids(self, current_data: Dict[str, Any], ext_id_name: str, ext_id_value: str,
                             update_mode: str) -> Tuple[Optional[Any], Dict[str, Dict[str, str]], bool]:
        """Internal helper function to prepare an update/addition of external
        id while preserving the others."""
        old_value = None
        if "externalIds" not in current_data:
            ext_id_data = {"externalIds": {ext_id_name: ext_id_value}}
        else:
            if ext_id_name in current_data["externalIds"]:
                old_value = current_data["externalIds"][ext_id_name]
            ext_id_data = {"externalIds": current_data["externalIds"]}
            ext_id_data["externalIds"][ext_id_name] = ext_id_value

        if update_mode == "delete":
            del ext_id_data["externalIds"][ext_id_name]

        if (update_mode == "overwrite"
                or (update_mode == "none" and old_value is None)
                or (update_mode == "delete" and old_value is not None)):
            update = True
        else:
            update = False

        return (old_value, ext_id_data, update)

    def _add_params(self, url: str, params: Dict[str, str]) -> str:
        """Add the given parameter to the given url"""

        query_string = urlencode(params)

        if "?" in url:
            return f"{url}&{query_string}"
        else:
            return f"{url}?{query_string}"

    @classmethod
    def get_id_from_href(cls, href: str) -> str:
        """"Extracts the identifier from the href and returns it

        :param href: HAL href for a specific resource
        :type href: string (valid URL)
        :return: the id part of the href
        :rtype: string
        """

        pos = href.rfind("/")
        identifier = href[(pos + 1):]
        return identifier



# -------------------------------------------------------------------------------
# Copyright (c) 2019-2025 Siemens
# Copyright (c) 2022 BMW CarIT GmbH
# All Rights Reserved.
# Authors: thomas.graf@siemens.com, gernot.hillier@siemens.com
# Authors: helio.chissini-de-castro@bmw.de
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

from typing import Any, Dict, List, Optional

from .base import BaseMixin
from .sw360error import SW360Error


class ReleasesMixin(BaseMixin):
    def get_release(self, release_id: str) -> Optional[Dict[str, Any]]:
        """Get information of about a release

        API endpoint: GET /releases/{id}

        :param release_id: the id of the release to be requested
        :type release_id: string
        :return: a release
        :rtype: JSON release object
        :raises SW360Error: if there is a negative HTTP response
        """
        resp = self.api_get(self.url + "resource/api/releases/" + release_id)
        return resp

    def get_release_by_url(self, release_url: str) -> Optional[Dict[str, Any]]:
        """Get information of about a release

        API endpoint: GET /releases

        :param url: the full url of the release to be requested
        :type url: string
        :return: a release
        :rtype: JSON release object
        :raises SW360Error: if there is a negative HTTP response
        """
        resp = self.api_get(release_url)
        return resp

    def get_releases_by_name(self, name: str) -> List[Any]:
        """Gets a list of releases that match the given name.

        API endpoint: GET /releases?name=

        :param name: the name
        :type name: string
        :return: list of releases
        :rtype: list of JSON release objects
        :raises SW360Error: if there is a negative HTTP response
        """
        full_url = self.url + "resource/api/releases?name=" + name
        resp = self.api_get(full_url)
        if resp and ("_embedded" in resp) and ("sw360:releases" in resp["_embedded"]):
            return resp["_embedded"]["sw360:releases"]

        return []

    # return type List[Dict[str, Any]] | Optional[Dict[str, Any]] for Python 3.11 is good,
    # Union[List[Dict[str, Any]], Optional[Dict[str, Any]]] for lower Python versions is not good
    def get_all_releases(self, fields: str = "", all_details: bool = False,
                         isNewClearingWithSourceAvailable: bool = False, page: int = -1,
                         page_size: int = -1, sort: str = "") -> Any:
        """Get information of about all releases

        API endpoint: GET /releases

        :param all_details: retrieve all project details (optional))
        :type all_details: bool
        :param isNewClearingWithSourceAvailable: retrieve releases in new clearning state with source avail
        :type isNewClearingWithSourceAvailable: bool
        :param page: page to retrieve
        :type page: int
        :param page_size: page size to use
        :type page_size: int
        :param sort: sort order for the releases ("name,desc"; "name,asc")
        :type sort: str
        :return: list of releases
        :rtype: list of JSON release objects
        :raises SW360Error: if there is a negative HTTP response
        """
        fullbase_url = self.url + "resource/api/releases"
        params = {}

        if all_details:
            params["allDetails"] = "true"

        if isNewClearingWithSourceAvailable:
            params["isNewClearingWithSourceAvailable"] = "true"

        if fields:
            params["fields"] = fields

        if page > -1:
            params["page"] = str(page)
            params["page_entries"] = str(page_size)

        if sort:
            params["sort"] = sort

        full_url = self._add_params(fullbase_url, params)

        resp = self.api_get(full_url)

        if page == -1 and resp and ("_embedded" in resp) and ("sw360:releases" in resp["_embedded"]):
            return resp["_embedded"]["sw360:releases"]

        return resp

    def get_releases_by_external_id(self, ext_id_name: str, ext_id_value: str = "") -> List[Dict[str, Any]]:
        """Get releases by external id. `ext_id_value` can be left blank to
        search for all releases with `ext_id_name`.

        API endpoint: GET /releases

        :param ext_id_name: the name of the external id to look for
        :param ext_id_value: the value of the external id to look for
        :type ext_id_name: string
        :type ext_id_value: string
        :return: list of releases
        :rtype: list of JSON release objects
        :raises SW360Error: if there is a negative HTTP response
        """
        resp = self.api_get(
            self.url
            + "resource/api/releases/searchByExternalIds?"
            + ext_id_name + "=" + ext_id_value
        )
        if resp and ("_embedded" in resp) and ("sw360:releases" in resp["_embedded"]):
            return resp["_embedded"]["sw360:releases"]

        return []

    def create_new_release(self, name: str, version: str, component_id: str,
                           release_details: Dict[str, Any] = {}) -> Optional[Dict[str, Any]]:
        """Create a new release

        API endpoint: POST /releases

        :param name: name of new release (usually set to component name)
        :param version: version string of new release (e.g. "1.0")
        :param component_id: SW360 ID of component in which release shall be created
        :param release_details: further release details as defined by SW360 REST API
        :type name: string
        :type version: string
        :type component_id: string
        :type release_details: dict
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        for param in "name", "version":
            release_details[param] = locals()[param]
        release_details["componentId"] = component_id

        url = self.url + "resource/api/releases"
        response = self.api_post(
            url, json=release_details)
        if response is not None:
            if response.ok:
                return response.json()
        return None

    def update_release(self, release: Dict[str, Any], release_id: str) -> Optional[Dict[str, Any]]:
        """Update an existing release

        API endpoint: PATCH /releases

        :param release: the new release data
        :param release_id: the id of the release to be deleted
        :type release: JSON
        :type release_id: string
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        if not release_id:
            raise SW360Error(message="No release id provided!")

        url = self.url + "resource/api/releases/" + release_id
        return self.api_patch(url, json=release)

    def update_release_external_id(self, ext_id_name: str, ext_id_value: str,
                                   release_id: str, update_mode: str = "none") -> Optional[Dict[str, Any]]:
        """Set or update external id of a release. If the id is already set, it
        will only be changed if `update_mode=="overwrite"`. The id can be
        deleted using `update_mode=="delete"`.

        The method will return the old value of the external id or None if it
        was not set.

        API endpoint: PATCH /releases

        :param ext_id_name: name of the external id
        :param ext_id_value: value of the external id
        :param release_id: the id of the release to be updated
        :param update_mode: can be "none" (default), "overwrite" or "delete"
        :type ext_id_name: string
        :type ext_id_value: string
        :type release_id: string
        :type update_mode: string
        :return: old value of external id
        :rtype: string
        :raises SW360Error: if there is a negative HTTP response
        """
        complete_data = self.get_release(release_id)
        if complete_data:
            ret = self._update_external_ids(complete_data, ext_id_name,
                                            ext_id_value, update_mode)
            (old_value, data, update) = ret
            if update:
                self.update_release(data, release_id)
            return old_value

        return {}

    def delete_release(self, release_id: str) -> Optional[Dict[str, Any]]:
        """Delete an existing release

        API endpoint: DELETE /releases

        :param release_id: the id of the release to be deleted
        :type release_id: string
        :return: SW360 result
        :rtype: JSON SW360 result object
        :raises SW360Error: if there is a negative HTTP response
        """

        if not release_id:
            raise SW360Error(message="No release id provided!")

        url = self.url + "resource/api/releases/" + release_id
        response = self.api_delete(url)
        if response is not None:
            if response.ok:
                if response.text:
                    return response.json()
        return None

    def get_users_of_release(self, release_id: str) -> Optional[Dict[str, Any]]:
        """Get information of about the users of a release

        API endpoint: GET /releases/usedBy/{id}

        :param release_id: the id of the release to be requested
        :type release_id: string
        :return: all users of this release
        :rtype: JSON objects
        :raises SW360Error: if there is a negative HTTP response
        """

        resp = self.api_get(self.url + "resource/api/releases/usedBy/" + release_id)
        return resp

    def link_packages_to_release(self, release_id: str, packages: List[str]) -> Optional[Dict[str, Any]]:
        """Link (new) packages to a given release.

        API endpoint PATCH /release/{pid}/packages{rid}

        :param release_id: the id of the existing release
        :type release_id: string
        :param packages: list of package ids
        :type packages: list of string
        :rtype: JSON SW360 result object
        :raises SW360Error: if the release id is missing ir there is a negative HTTP response
        """
        if not release_id:
            raise SW360Error(message="No release id provided!")

        url = self.url + "resource/api/releases/" + release_id + "/link/packages/"
        return self.api_patch(url, json=packages)

    def unlink_packages_from_release(self, release_id: str, packages: List[str]) -> Optional[Dict[str, Any]]:
        """Unlink packages from a given release.

        API endpoint PATCH /release/{pid}/packages{rid}

        :param release_id: the id of the existing release
        :type release_id: string
        :param packages: list of package ids
        :type packages: list of string
        :rtype: JSON SW360 result object
        :raises SW360Error: if the project id is missing ir there is a negative HTTP response
        """
        if not release_id:
            raise SW360Error(message="No release id provided!")

        url = self.url + "resource/api/releases/" + release_id + "/unlink/packages/"
        return self.api_patch(url, json=packages)

    def get_recent_releases(self) -> Optional[List[Dict[str, Any]]]:
        """Get 5 of the service's most recently created releases.

        API endpoint: GET /releases/recentReleases

        :return: a list of releases
        :rtype: JSON list of release objects
        :raises SW360Error: if there is a negative HTTP response
        """
        url = self.url + "resource/api/releases/recentReleases"
        resp = self.api_get(url)
        if resp and ("_embedded" in resp) and ("sw360:releases" in resp["_embedded"]):
            return resp["_embedded"]["sw360:releases"]

        return []



# -------------------------------------------------------------------------------
# Copyright (c) 2019-2023 Siemens
# Copyright (c) 2022 BMW CarIT GmbH
# All Rights Reserved.
# Authors: thomas.graf@siemens.com, gernot.hillier@siemens.com
# Authors: helio.chissini-de-castro@bmw.de
#
# Licensed under the MIT license.
# SPDX-License-Identifier: MIT
# -------------------------------------------------------------------------------

import json
import logging
import os
from http import HTTPStatus
from typing import Any, Dict, List, Optional

import requests

from .base import BaseMixin
from .sw360error import SW360Error

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class AttachmentsMixin(BaseMixin):
    def get_attachment_infos_by_hash(self, hashvalue: str) -> Optional[Dict[str, Any]]:
        """Get information about attachments with a given sha1 hash value.

        This usually returns zero or one result, but if the same binary file
        was uploaded and attached to multiple sw360 resources, it will return
        all matching attachments for the given hash.

        API endpoint: GET /attachments?sha1=
        """

        resp = self.api_get(
            self.url + "resource/api/attachments?sha1=" + hashvalue
        )
        return resp

    def get_attachment_infos_for_resource(self, resource_type: str, resource_id: str) -> List[Dict[str, Any]]:
        """Get information about the attachments of a specific resource.

        Usually, you don't need to call this directly, but use one of the
        specific get_attachment_infos_for_{release,component,project} functions.
        """

        resp = self.api_get(
            self.url + "resource/api/"
            + resource_type
            + "/"
            + resource_id
            + "/attachments"
        )

        if resp and "_embedded" in resp and "sw360:attachments" in resp["_embedded"]:
            return resp["_embedded"]["sw360:attachments"]

        return []

    def get_attachment_infos_for_release(self, release_id: str) -> List[Dict[str, Any]]:
        """Get information about the attachments of a release

        :param release_id: id of the release from which to list attachments
        """

        resp = self.get_attachment_infos_for_resource("releases", release_id)
        return resp

    def get_attachment_infos_for_component(self, component_id: str) -> List[Dict[str, Any]]:
        """Get information about the attachments of a component

        :param component_id: id of the component from which to list attachments
        """

        resp = self.get_attachment_infos_for_resource("components", component_id)
        return resp

    def get_attachment_infos_for_project(self, project_id: str) -> List[Dict[str, Any]]:
        """Get information about the attachments of a project

        :param project_id: id of the project from which to list attachments
        """

        resp = self.get_attachment_infos_for_resource("projects", project_id)
        return resp

    def get_attachment_by_url(self, url: str) -> Optional[Dict[str, Any]]:
        """Get information about attachment.

        :param url: the full url of the attachment to be requested
        """
        resp = self.api_get(url)
        return resp

    def get_attachment(self, attachment_id: str) -> Optional[Dict[str, Any]]:
        """Get information about an attachment

        API endpoint: GET /attachments

        :param attachment_id: id of the attachment
        """

        resp = self.api_get(self.url + "resource/api/attachments/" + attachment_id)
        return resp

    def download_release_attachment(self, filename: str, release_id: str, attachment_id: str) -> None:
        """Downloads an attachment of a release

        API endpoint: GET /attachments
        """

        self.download_resource_attachment(
            filename, "releases", release_id, attachment_id
        )

    def download_project_attachment(self, filename: str, project_id: str, attachment_id: str) -> None:
        """Downloads an attachment of a project

        API endpoint: GET /attachments
        """

        self.download_resource_attachment(
            filename, "projects", project_id, attachment_id
        )

    def download_component_attachment(self, filename: str, component_id: str, attachment_id: str) -> None:
        """Downloads an attachment of a component

        API endpoint: GET /attachments
        """

        self.download_resource_attachment(
            filename, "components", component_id, attachment_id
        )

    def download_resource_attachment(self, filename: str, resource_type: str, resource_id: str,
                                     attachment_id: str) -> None:
        """Downloads an attachment from SW360 (only for internal use)

        API endpoint: GET /attachments
        """

        if not resource_id:
            raise SW360Error(message="No resource id provided!")

        if not attachment_id:
            raise SW360Error(message="No attachment id provided!")

        url = (
            self.url
            + "resource/api/"
            + resource_type
            + "/"
            + resource_id
            + "/attachments/"
            + attachment_id
        )
        self.download_attachment(filename, url)

    def download_attachment(self, filename: str, download_url: str) -> None:
        """Downloads an attachment from SW360

        API endpoint: GET /attachments
        """

        hdr = self.api_headers.copy()
        hdr["Accept"] = "application/*"
        req = requests.get(download_url, allow_redirects=True, headers=hdr)
        if req.ok:
            open(filename, "wb").write(req.content)
        else:
            raise SW360Error(req, download_url)

    def _upload_resource_attachment(self, resource_type: str, resource_id: str, upload_file: str,
                                    upload_type: str = "SOURCE", upload_comment: str = "") -> None:
        """Upload `upload_file` as attachment to SW360 for the resource with the given id
        using `upload_comment` for it. `upload_type` can be
        "DOCUMENT"
        "SOURCE"
        "CLEARING_REPORT"
        "COMPONENT_LICENSE_INFO_XML"
        "SOURCE_SELF"
        "BINARY"
        "BINARY_SELF"
        "LICENSE_AGREEMENT"
        "README_OSS"

        API endpoint: POST /attachments
        """
        if not os.path.exists(upload_file):
            raise SW360Error(message="ERROR: file not found: " + upload_file)

        if resource_type not in ("releases", "components", "projects"):
            raise SW360Error(message="Invalid resource type provided!")

        if (not isinstance(resource_id, str)) or (resource_id == ""):
            raise SW360Error(message="Invalid resource id provided!")

        filename = os.path.basename(upload_file)
        url = self.url + "resource/api/" + resource_type + "/" + resource_id + "/attachments"
        attachment_data = {"filename": filename,
                           "attachmentContentId": "2",
                           "createdComment": upload_comment,
                           "attachmentType": upload_type}
        file_data = {
            "file": (filename, open(upload_file, "rb"), "multipart/form-data"),
            "attachment": (
                "",  # dummy filename
                json.dumps(attachment_data),
                "application/json",
            ),
        }
        response = self.api_post_multipart(url, files=file_data)
        if response is not None:
            if response.status_code == HTTPStatus.ACCEPTED:
                logger.warning(
                    f"Attachment upload was accepted by {url} but might not be visible yet: {response.text}"
                )
            if not response.ok:
                raise SW360Error(response, url)

    def upload_release_attachment(self, release_id: str, upload_file: str, upload_type: str = "SOURCE",
                                  upload_comment: str = "") -> None:
        """Upload `upload_file` as attachment to SW360 for `release_id`,
        using `upload_comment` for it. `upload_type` can be
        "DOCUMENT"
        "SOURCE"
        "CLEARING_REPORT"
        "COMPONENT_LICENSE_INFO_XML"
        "SOURCE_SELF"
        "BINARY"
        "BINARY_SELF"
        "LICENSE_AGREEMENT"
        "README_OSS"

        API endpoint: POST /attachments
        """

        self._upload_resource_attachment(
            "releases", release_id, upload_file,
            upload_type=upload_type, upload_comment=upload_comment)

    def upload_component_attachment(self, component_id: str, upload_file: str, upload_type: str = "SOURCE",
                                    upload_comment: str = "") -> None:
        """Upload `upload_file` as attachment to SW360 for `component_id`,
        using `upload_comment` for it. `upload_type` can be
        "DOCUMENT"
        "SOURCE"
        "CLEARING_REPORT"
        "COMPONENT_LICENSE_INFO_XML"
        "SOURCE_SELF"
        "BINARY"
        "BINARY_SELF"
        "LICENSE_AGREEMENT"
        "README_OSS"

        API endpoint: POST /attachments
        """

        self._upload_resource_attachment(
            "components", component_id, upload_file,
            upload_type=upload_type, upload_comment=upload_comment)

    def upload_project_attachment(self, project_id: str, upload_file: str, upload_type: str = "SOURCE",
                                  upload_comment: str = "") -> None:
        """Upload `upload_file` as attachment to SW360 for `project_id` of,
        using `upload_comment` for it. `upload_type` can be
        "DOCUMENT"
        "SOURCE"
        "CLEARING_REPORT"
        "COMPONENT_LICENSE_INFO_XML"
        "SOURCE_SELF"
        "BINARY"
        "BINARY_SELF"
        "LICENSE_AGREEMENT"
        "README_OSS"

        API endpoint: POST /attachments
        """

        self._upload_resource_attachment(
            "projects", project_id, upload_file,
            upload_type=upload_type, upload_comment=upload_comment)
